;; $Id: pv-mode.el,v 1.2 2001/10/14 15:55:41 jfine Exp $

;; pv-mode.el -- provide Instant Preview for (La)TeX files
;; Copyright 2001 by Jonathan Fine <jfine@activetex.org>

;; Instant Preview is free software distributed under the terms 
;; of the the GNU General Public License.

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License; either
;; version 2 of the License, or (at your option) any later version.

;; For more information about this program, send a message to
;; mailto:jfine@activetex.org

;; $Log: pv-mode.el,v $
;; Revision 1.2  2001/10/14 15:55:41  jfine
;; Revised for v0.2 of the package
;;
;; Revision 1.1  2001/05/27 19:55:54  jfine
;; Initial revision
;;

;; The book "Writing GNU Emacs Extensions" by Bob Glickstein
;; (pub: O'Reilly, First Edition) has been used as a source.

;; Preview is an unusual mode, in that by design it adds no new
;; editing functionality to Emacs.  Instead, it adds display
;; functionality.

;; When Preview is enabled, one gets instant preview of the (La)TeX
;; file one is editing. Except only LaTeX implemented so far.

;; You will need to have "texd" installed, to use this package.
;; This version of pv-mode.el is part of texd-0.2.  Please read the
;; installation instructions there.


;; the preview minor mode variable
(defvar pv-mode nil
  "Mode variable for preview minor mode.")

;; these variables of most interest to the end user
(defvar pv-belay-re "\n\\\\Belay{.*\n"
  "Regular expression to match preview belay.")

(defvar pv-header-re "^\\\\begin{document}.*$"
  "Regular expression to match header.")

(defvar pv-delim-re "^[ \t]*\n"
  "Regular expression to delimit preview chunks.")

;; these variables are buffer local
(make-variable-buffer-local 'pv-mode)
(make-variable-buffer-local 'pv-belay-re)
(make-variable-buffer-local 'pv-header-re)
(make-variable-buffer-local 'pv-delim-re)


;; working directory and dvi file to preview
(defconst pv-tmp-dir (format "/tmp/%s/pvtexd" (getenv "LOGNAME"))
  "Directory used to store pv-mode's temporary files")

(defconst pv-dvi-file (format "%s/%i.dvi" pv-tmp-dir (emacs-pid))
  "The dvi file displayed by the previewer (content is variable)" )


;; set up the preview minor mode
(defun pv-mode (&optional arg)
  "Preview minor mode."
  (interactive "P")
  (setq pv-mode
	(if (null arg)
	    (not pv-mode)
	  (> (prefix-numeric-value arg) 0)))

  (make-local-hook 'pre-command-hook)
  (make-local-hook 'window-scroll-functions)
  (make-local-hook 'after-change-functions)
  (make-local-hook 'kill-buffer-hook)

  (if pv-mode
      (progn
	(pv-start)
	(add-hook 'pre-command-hook 'pv-pre-command nil t)
	(add-hook 'window-scroll-functions 'pv-maybe-refresh nil t) 
	(add-hook 'after-change-functions 'pv-maybe-refresh nil t) 
	(add-hook 'kill-buffer-hook 'pv-end nil t))

    (remove-hook 'kill-buffer-hook 'pv-end t) 
    (remove-hook 'after-change-functions 'pv-maybe-refresh t)
    (remove-hook 'window-scroll-functions 'pv-maybe-refresh t)
    (remove-hook 'pre-command-hook 'pv-pre-command t)
    (pv-end))
  )

;; add "Preview" to the minor mode association list
(if (not (assq 'pv-mode minor-mode-alist))
    (setq minor-mode-alist
	  (cons `(pv-mode " Preview")
		minor-mode-alist)))


;; we will need a dvi preview process
(defvar pv-xdvi-process nil
  "If not nil, holds the preview process")
  
(defun start-xdvi()
  "Start xdvi (and restart it if it exist abnormally)"
  (copy-file "~/.texd/null.dvi"  pv-dvi-file t)
  (setq pv-xdvi-process 
	(start-process 
	 "xdvi" nil "xdvi" "-paper" "1ptx1pt" 
	 "-geometry" "650x650"
	 pv-dvi-file))
  (set-process-sentinel pv-xdvi-process `pv-xdvi-sentinel)
  t)

(defun pv-xdvi-sentinel (process string)
  "Restart xdvi if it exits abnormally"
  (if (zerop (process-exit-status process))
      (setq pv-xdvi-process nil)
    (start-xdvi)
    )
  )

;; we will need a tex process
(defvar pv-tex-process nil
  "If not nil, holds the tex process (buffer local)")

(make-variable-buffer-local 'pv-tex-process)

;; we need to start a TeX daemon for each buffer in pv-mode
(defun pv-start()
  "Initialisation code for pv-mode"
  (if (null (processp pv-xdvi-process))
      (start-xdvi))
  (pv-write-hdr)
  (setq pv-tex-process
	(start-process (format "tex %s" (buffer-name))
		       nil "xterm" "-e" "pvtexd" pv-tmp-dir "latex" (buffer-name)))
  (set-process-sentinel pv-tex-process `pv-tex-sentinel)
  (put 'pv-end pv-tex-process (current-buffer))
  )

(defun pv-tex-sentinel (process string)
  "If process ends, exit preview mode"
  (let ((buffer (get 'pv-end process)))
    (if (bufferp buffer)
	(progn
	  (set-buffer (get 'pv-end process))
	  (setq pv-tex-process nil)
	  (pv-mode -1)))))

(defun pv-end()
  "Cleanup code for pv-mode"
  (if pv-tex-process
      (kill-process pv-tex-process)
    (setq pv-tex-process nil)))

;;  a single command might both change and scroll the buffer
(defvar pv-maybe-refresh nil
  "Set to t before each command, and to nil after every refresh")

(defun pv-pre-command()
  "Set pv-maybe-refresh to t"
  (setq pv-maybe-refresh t))


(defun pv-maybe-refresh(&rest unused)
  "If appropriate and possible, refresh the preview"
  (if pv-maybe-refresh
      (progn
	(pv)
	(setq pv-maybe-refresh nil))))


(defun pv-write-hdr()
  "Write the buffer's (La)TeX header out for use by TeX daemon"
  (save-excursion
    (save-restriction
      (save-match-data
	(widen)
	(goto-char (point-min))
	(if (re-search-forward pv-header-re 10000 t)
	    (progn
	      (write-region (point-min) (point) 
			  (format "%s/%s.hdr" pv-tmp-dir (buffer-name))
			  nil 'no-message)
	      t))))))


;; start the tex daemon
(defun pv-start-texd ()
  "Start a TeX daemon associated with the current buffer"
  (make-comint 
   (format "%s.log" (buffer-name))
   "pvtexd" nil pv-tmp-dir "latex" (buffer-name)))


;; this command finds a region to preview, and sends it to the TeX daemon
;; it could do better, when it comes to finding the region
;; (would anyone like to take this on?)
(defun pv(&rest unused)
  "Preview chunk containing point"
  (if pv-mode
      (save-excursion
	(save-match-data
	  (let (belay-locn belay start end)
	    (move-to-window-line 0)
	    (setq start (re-search-backward pv-delim-re nil t nil))
	    (move-to-window-line 0)
	    (if (setq belay-locn (re-search-backward pv-belay-re nil t nil))
		(setq belay (match-string 0))
	      (setq belay "\\Belay{0}\n"))
	    (move-to-window-line -1)
	    (setq end (re-search-forward pv-delim-re nil t nil))
	    (if (and belay start end 
		     (= 0 (call-process "rm" nil nil nil 
					(format "%s/%s.key" pv-tmp-dir (buffer-name)))))
		(pv-belayed-region belay start end)
	      (message "Unable to preview chunk")
	      )
	    )
	  )
	)
    (message "Buffer not in preview mode.")
    nil
    )
  )


;; this helper command sends material out to the TeX daemon
(defun pv-belayed-region (belay start end)
  "Using BELAY preview from START to END."
  (if (null (processp pv-xdvi-process))
      (start-xdvi))
  (write-region (concat 
		 (format "\\global\\mathchardef\\ChopPid %i\n" (process-id pv-xdvi-process))
		 (format "\\global\\mathchardef\\ChopNum %i\n" (emacs-pid))
		 belay)
		nil
		(format "%s/%s.cfg" pv-tmp-dir (buffer-name))
		nil 'no-message)
  (write-region start end (format "%s/%s.in" pv-tmp-dir (buffer-name))
		nil 'no-message)
)

(provide 'pv-mode)

;; pv.el ends here
